@@ -1,5 +1,6 @@
 import csv
 import json
+import pdb
 import requests
 import time
 
@@ -10,59 +11,96 @@ class IterableApi():
 	"""
 	This is a python wrapper for the Iterable API
 
-	We are using the Requests HTTP python library, which I 
-	have found flexible in regards to the arguments you
-	can pass into each request.  Their documentation is 
-	also excellent.  This makes it must easier in the event I were to 
-	expand this wrapper to encompass all the possible Iterable API requests.
+	We are using the 'Requests' HTTP python library, which I 
+	have found very flexible to accomodate the various methods
+	that customers leverage to interact with our API.  'Requests' 
+	documentation is also excellent, enabling our team to 
+	quickly update this wrapper to support a wide range of use cases.  
 
 	"""	
 
 	def __init__(self, api_key):
 		"""
-		This preforms all initialization and stores the unique API key of the
-		Iterable instance. It also stores the base URI, which is consistent
-		for all instances.
+		This preforms the necessary initialization parameters for the
+		Iterable API wrapper. It stores the base URI, the API key for the 
+		project, and headers that shoudl be consistent across all requests. 
 
 		"""
-		
-		self.base_uri = "https://api.iterable.com"		
 		self.api_key = api_key
+		self.base_uri = "https://api.iterable.com"
+		self.headers = {
+						"Content-type": "application/json",
+						"Api-Key": self.api_key
+						}		
 
 	def api_call(self, call, method, params=None, headers=None, data=None,
 				 json=None):
 		"""
-		This is our generic api call function.  We will route all our calls
-		through this function.  The benefit of this is that it:
+		This is our generic api call function.  We will route all calls except
+		requests that do not return JSON ('Export' and 'Experiment Metrics' are
+		examples where this is the case).  This is beneficial because:
 			1. Allows for easier debugging if a request fails
-			2. Even though the Iterable API only needs the API key for a 
-			security standpoint, if it were ever to require access token for 
-			each request we could easily manage the granting and expiration
-			management of such a token.  
+			2. Currently, Iterable only needs the API key from a security
+			standpoint. In the future, if it were to require an  
+			access token for each request we could easily manage the granting
+			and expiration management of such a token.  
 
 		"""
 
 		# params(optional) Dictionary or bytes to be sent in the query string for the Request.
 		if params is None:
 			params = {}
-		# headers- dictionary of HTTP Headers to be sent with Request
-		if headers is None:
-			headers = {}
 		# data- dict or list of tuples to be sent in body of Request
 		if data is None:
 			data = {}
 		# json- data to be sent in body of Request
 		if json is None:
-			json ={}		
-		
-		headers["Content-type"] = "application/json"
-		headers["Api-Key"] = self.api_key
+			json ={}
 
 		# make the request following the 'requests.request' method
 		r = requests.request(method=method, url=self.base_uri+call, params=params,
-							 headers=headers, data=data, json=json)
+							 headers=self.headers, data=data, json=json)	
 
-		return r.json()
+		response = {			
+			"body": r.json(),			
+			"code": r.status_code,
+			"headers": r.headers,
+			"url": r.url
+		}
+
+		return response
+
+	def export_data_api(self, call,
+						params, path, 
+						chunk_size=None, 
+						return_response_object=None):
+
+		r = requests.request(method="GET", url=self.base_uri+call, params=params,
+							 headers=self.headers, stream=True)
+
+		if r.status_code == 200:
+			
+			if return_response_object is (not None and True):
+
+				return r
+
+
+			if "csv" in r.url:
+				local_filename = 'iterable_' + params['dataTypeName'] + str(round(time.time())) + '.csv'
+			if "experiments" in r.url:
+				local_filename = 'iterable_experiment_ids_' + str(",".join(list(params.values()))) + "_" + str(round(time.time())) + '.csv'
+			if "userEvents" in r.url:
+				local_filename = 'iterable_user_events' + str(round(time.time())) + '.csv'
+			if "json" in r.url:
+				local_filename = 'iterable_' + params['dataTypeName'] + str(round(time.time())) + '.json'
+
+			with open(path+local_filename, 'wb') as write_file:
+				
+				for chunk in r.iter_content(chunk_size=chunk_size):
+					if chunk:
+						write_file.write(chunk)
+			
+			return True
 
 	"""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
 
@@ -70,7 +108,7 @@ class IterableApi():
 
 	"""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
 
-	def list_campaign_info(self):
+	def list_campaign_metadata(self):
 
 		call="/api/campaigns"
 
@@ -159,7 +197,7 @@ class IterableApi():
 
 	"""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
 
-	def list_channels(self):
+	def get_channels(self):
 
 		call="/api/channels"
 
@@ -171,9 +209,14 @@ class IterableApi():
 
 	"""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
 
-	def track_purchase(self, user, items, total, campaign_id=None, 
+	def track_purchase(self, user, items, total, purchase_id= None, campaign_id=None, 
 					   template_id=None, created_at=None,
 					   data_fields=None):
+		"""
+			The 'purchase_id' argument maps to 'id' for this API endpoint.
+			This name is used to distinguish it from other instances where
+			'id' is a part of the API request with other Iterable endpoints.
+		"""
 
 		call="/api/commerce/trackPurchase"
 
@@ -194,6 +237,9 @@ class IterableApi():
 		else:
 			raise TypeError('total is not in correct format')
 
+		if purchase_id is not None:
+			payload["id"]= str(purchase_id) 
+
 		if campaign_id is not None:
 			payload["campaignId"]= campaign_id
 
@@ -288,7 +334,7 @@ class IterableApi():
 
 		payload={}
 
-		if (limit is not None and limit <= 200):
+		if limit is not None and limit <= 200:
 			payload["limit"]= limit
 
 		return self.api_call(call=call, method="GET", params=payload)
@@ -313,9 +359,9 @@ class IterableApi():
 
 		return self.api_call(call=call, method="POST", json=payload)
 
-	def track_event(self, event_name, email=None, created_at=None,
-					data_fields=None, user_id=None, campaign_id=None,
-					template_id=None):
+	def track_event(self, event_name, event_id=None, email=None, 
+					created_at=None, data_fields=None, user_id=None,
+					campaign_id=None,template_id=None):
 
 		call="/api/events/track"
 
@@ -323,6 +369,9 @@ class IterableApi():
 
 		payload["eventName"]= str(event_name)
 
+		if event_id is not None:
+			payload["id"]= str(event_id)
+
 		if email is not None:
 			payload["email"]=email
 
@@ -383,7 +432,7 @@ class IterableApi():
 
 		return self.api_call(call=call, method="POST", json=payload)
 
-	def track_push_open(self,  campaign_id,email=None, user_id=None,
+	def track_push_open(self, campaign_id, email=None, user_id=None,
 						template_id=None, message_id=None, created_at=None,
 						data_fields=None):
 
@@ -443,11 +492,28 @@ class IterableApi():
 
 	"""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
 
-	def get_experiment_metrics(self, experiment_id=None, campaign_id=None,
-							   start_date_time=None, end_date_time=None):
+	def get_experiment_metrics(self, path, return_response_object= None,
+							   experiment_id=None, campaign_id=None,
+							   start_date_time=None, end_date_time=None
+							   ):
+		"""
+			This endpoint doesn't return a JSON object, instead it returns
+			a series of rows, each its own object. Given this setup, it makes 
+			sense to treat it how we handle our Bulk Export reqeusts.
+
+			Arguments:
+
+			path: the directory on your computer you wish the file to be downloaded into.
+			
+			return_response_object: recommended to be set to 'False'.  If set to 'True', 
+			will just return the response object as defined by the 'python-requests' module.
+			"""
 
 		call="/api/experiments/metrics"
 
+		if isinstance(return_response_object, bool) is False:
+			raise ValueError("'return_iterator_object'parameter must be a boolean") 
+
 		payload={}
 
 		if experiment_id is not None:
@@ -462,18 +528,19 @@ class IterableApi():
 		if end_date_time is not None:
 			payload["endDateTime"]=end_date_time
 
-		return self.api_call(call=call, method="GET", params=payload)
+		return self.export_data_api(call=call, path=path, params=payload)
 
 	"""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
 	
-	Export Data Requests
+	Export Requests
 
 	"""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
 
 	def export_data_csv(self, data_type_name=None, date_range=None,
 						delimiter=None, start_date_time=None,
 						end_date_time=None, omit_fields=None,
-						only_fields=None, campaign_id=None):
+						only_fields=None, campaign_id=None,
+						path=None):
 
 		call="/api/export/data.csv"
 
@@ -495,7 +562,7 @@ class IterableApi():
 			payload["endDateTime"]= end_date_time
 
 		if omit_fields is not None:
-			payload["omitFields"]= omitFields
+			payload["omitFields"]= omit_fields
 
 		if only_fields is not None and isinstance(only_fields, list):
 			payload["onlyFields"]= only_fields
@@ -503,26 +570,56 @@ class IterableApi():
 		if campaign_id is not None:
 			payload["campaignId"]= campaign_id
 
-		return self.api_call(call=call, method="GET", params=payload)
+		return self.export_data_api(call=call, params=payload, path=path)
 
-	def export_data_json(self, data_type_name=None, date_range=None,
+	def export_data_json(self, return_response_object, 
+						chunk_size=1024, 
+						path=None,
+						data_type_name=None, date_range=None,
 						delimiter=None, start_date_time=None,
 						end_date_time=None, omit_fields=None,
 						only_fields=None, campaign_id=None):
 
+		"""
+		Custom Keyword arguments:
+
+		1. return_response_object:
+			if set to 'True', the 'r' response object will be returned.  The
+			benefit of this is that you can manipulate the data in any way you
+			want.  If set to false, we will write the response to a file where each
+			Iterable activity you're exporting is a single-line JSON object.
+		2. chunk_size:
+			Chunk size is used as a paremeter in the r.iter_content(chunk_size) method
+			that controls how big the response chunks are (in bytes).  Depending on the
+			device used to make the request, this might change depending on the user. 
+			Default is set to 1 MB. 
+		3. path:
+			Allows you to choose the directory where the file is downloaded into.
+				Example: "/Users/username/Desktop/"
+			If not set the file will download into the current directory.
+			
+		"""
 		call="/api/export/data.json"
 
+		# make sure correct ranges are being used
+		date_ranges = ["Today", "Yesterday", "BeforeToday", "All"]		
+		
+		if isinstance(return_response_object, bool) is False:
+			raise ValueError("'return_iterator_object'parameter must be a boolean") 
+		
+		if chunk_size is not None and isinstance(chunk_size, int):
+			pass
+		else:
+			raise ValueError("'chunk_size' parameter must be a integer")
+
 		payload={}
 
 		if data_type_name is not None:
 			payload["dataTypeName"]= data_type_name
 
-		if date_range is not None:
+		if date_range is not None and date_range in date_ranges:
 			payload["range"]= date_range
 
-		if delimiter is not None:
-			payload["delimiter"]= delimiter
-
 		if start_date_time is not None:
 			payload["startDateTime"]= start_date_time
 
@@ -530,7 +627,7 @@ class IterableApi():
 			payload["endDateTime"]= end_date_time
 
 		if omit_fields is not None:
-			payload["omitFields"]= omitFields
+			payload["omitFields"]= omit_fields
 
 		if only_fields is not None and isinstance(only_fields, list):
 			payload["onlyFields"]= only_fields
@@ -538,7 +635,26 @@ class IterableApi():
 		if campaign_id is not None:
 			payload["campaignId"]= campaign_id
 
-		return self.api_call(call=call, method="GET", params=payload)
+		return self.export_data_api(call=call, chunk_size=chunk_size, 
+									params=payload, path=path,
+									return_response_object=return_response_object)
+
+
+	def export_user_events(self, email, include_custom_events,
+						   path, return_response_object= None):
+
+		call ="/api/export/userEvents"
+
+		if isinstance(include_custom_events, bool) is False:
+			raise ValueError("'include_custom_events' parameter must be a boolean")
+
+		payload = {}
+
+		payload["email"]= str(email)
+
+		payload["includeCustomEvents"] = include_custom_events
+
+		return self.export_data_api(call=call, params=payload, path=path)
 
 	"""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
 	
@@ -610,7 +726,7 @@ class IterableApi():
 
 		return self.api_call(call=call, method="GET")
 
-	def create_list(self, list_name):
+	def create_static_list(self, list_name):
 
 		call = "/api/lists"
 
@@ -626,7 +742,7 @@ class IterableApi():
 
 		return self.api_call(call=call, method="DELETE")
 
-	def number_of_users_in_list(self, list_id):
+	def count_of_users_in_list(self, list_id):
 
 		call = "/api/lists/"+str(list_id)+"/size"
 
@@ -830,7 +946,7 @@ class IterableApi():
 
 	"""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
 
-	def get_templates_for_project(self, template_type=None, 
+	def get_templates(self, template_type=None, 
 								  message_medium=None,
 								  start_date_time=None,
 								  end_date_time=None):
@@ -839,27 +955,31 @@ class IterableApi():
 
 		payload={}
 
-		if template_type is not None and template_type is not ("Base" or "Blast" or "Triggered" or "Workflow"):
-				
-			raise ValueError('You did not specify the correct template type')
-		else:
-			payload["templateType"]=template_type
+		iterable_template_types = ["Base", "Blast", "Triggered", "Workflow"]
 
-		if message_medium is not None and message_medium is not ("Email" or "Push" or "InApp" or "SMS"):
-			
-			raise ValueError('You did not specify the correct message medium')
-		else:			
-			payload["messageMedium"]=message_medium
+		iterable_message_mediums = ["Email", "Push", "InApp", "SMS"]
+
+		if template_type is not None and template_type in iterable_template_types:			
+			payload["templateType"]= template_type
+
+		elif template_type is not None and template_type not in iterable_template_types:
+			raise Exception("It looks like you listed an incorrect template type '%s'" % template_type)	
+
+		if message_medium is not None and message_medium in iterable_message_mediums:			
+			payload["messageMedium"]= message_medium		
+
+		elif message_medium is not None and message_medium not in iterable_message_mediums:
+			raise Exception("It looks like you listed an incorrect message medium '%s'" % message_medium)
 			
 		if start_date_time is not None:
-			payload["startDateTime"]=start_date_time
+			payload["startDateTime"]= start_date_time
 
 		if end_date_time is not None:
 			payload["endDateTime"]= end_date_time
 
-		return self.api_call(call=call, method="GET", params=payload)
+		return self.api_call(call = call, method = "GET", params = payload)
 
-	def get_email_template(self, template_id, locale=None):
+	def get_email_template_by_templateId(self, template_id, locale=None):
 
 		call="/api/templates/email/get"
 
@@ -1028,7 +1148,7 @@ class IterableApi():
 
 		return self.api_call(call=call, method="POST", json=payload)
 
-	def get_email_template(self, client_template_id):
+	def get_email_template_by_client_templateId(self, client_template_id):
 
 		call="/api/templates/getClientTemplateId"
 
@@ -1235,7 +1355,7 @@ class IterableApi():
 
 	"""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
 
-	def delete_user(self, email):
+	def delete_user_by_email(self, email):
 	
 		"""
 		This call will delete a user from the Iterable database.  
@@ -1353,12 +1473,6 @@ class IterableApi():
 						  campaign_id=None, start_date_time=None,
 						  end_date_time=None, exclude_blast_campaigns=None,
 						  message_medium=None):
-		"""
-
-		campaign_id takes an Array[double] as a query parameter...what
-		does this mean x=[[1,2,3],[4,5,6]]?
-
-		"""
 
 		call = "/api/users/getSentMessages"
 
@@ -1427,7 +1541,7 @@ class IterableApi():
 		return self.api_call(call=call, method="POST", json=payload)
 
 	def update_user(self, email=None, data_fields=None, user_id=None,
-					merge_nested_objects=None):
+					prefer_userId= None, merge_nested_objects=None):
 
 		"""
 		The Iterable 'User Update' api updates a user profile with new data 
@@ -1457,21 +1571,32 @@ class IterableApi():
 		if user_id is not None:
 			payload["userId"] = str(user_id)
 
+		if prefer_userId is not None:
+			payload["preferUserId"]= prefer_userId
+
 		if merge_nested_objects is not None:
 			payload["mergeNestedObjects"] = merge_nested_objects
 		
 		return self.api_call(call=call, method="POST", json=payload)
 
-	def update_email(self, current_email, new_email):
+	def update_email(self, new_email, current_email= None,
+					 current_userid= None):
 
 		call = "/api/users/updateEmail"
 
 		payload = {}
 
-		payload["currentEmail"] = str(current_email)
+		if current_email is not None:
+			payload["currentEmail"] = str(current_email)
+
+		if current_userid is not None:
+			payload["currentUserId"] = str(current_userid)
 
 		payload["newEmail"] = str(new_email)
 
+		if ('currentEmail' or 'currentUserId') in payload == False:
+			raise ValueError('You need to pass in either email or username into the function')
+
 		return self.api_call(call=call, method="POST", json=payload)
 
 	def update_subscriptions(self, email, email_list_ids=None,
@@ -1502,6 +1627,32 @@ class IterableApi():
 
 		return self.api_call(call=call, method="POST", json=payload)
 
+	def get_forgotten_users_in_complaince_with_gdpr(self):
+
+		call= "/api/users/forgotten"
+
+		return self.api_call(call=call, method="GET")
+
+	def forget_a_user_in_compliance_with_gdpr(self, email):
+
+		call= "/api/users/forget"
+
+		payload ={}
+
+		payload["email"]= email
+
+		return self.api_call(call=call, method="POST", json=payload)
+
+	def unforget_a_user_in_compliance_with_gdpr(self, email):
+
+		call="/api/users/unforget"
+
+		payload={}
+
+		payload["email"]= email
+
+		return self.api_call(call=call, method="POST", json=payload)
+
 	"""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
 
 	Iterable Web Push Requests
@@ -1563,6 +1714,7 @@ class IterableApi():
 
 		return self.api_call(call=call, method="POST", json=payload)
 
+	
 
 
 
